

from typing import Any, Callable, Dict, Optional

from mitosheet.public.v3.types.bool import cast_to_bool
from mitosheet.public.v3.types.datetime import cast_to_datetime
from mitosheet.public.v3.types.float import cast_to_float
from mitosheet.public.v3.types.int import cast_to_int
from mitosheet.public.v3.types.number import cast_to_number
from mitosheet.public.v3.types.str import cast_to_string
from mitosheet.public.v3.types.timedelta import cast_to_timedelta

from mitosheet.types import PrimitiveTypeName

CONVERSION_FUNCTIONS: Dict[PrimitiveTypeName, Callable[[Any], Optional[Any]]] = {
    'str': cast_to_string,
    'int': cast_to_int,
    'float': cast_to_float,
    'number': cast_to_number,
    'bool': cast_to_bool,
    'datetime': cast_to_datetime,
    'timedelta': cast_to_timedelta,
}




from typing import Dict, Optional, Tuple, Union

from openpyxl import load_workbook

from mitosheet.excel_utils import get_col_and_row_indexes_from_range, get_column_from_column_index


def get_table_range_from_upper_left_corner_value(file_path: str, sheet_name: str, upper_left_value: Union[str, int, float, bool], bottom_left_value: Optional[Union[str, int, float, bool]]=None) -> Optional[str]:
    """
    Given a string, this function will look through the excel tab sheet_name at the given
    file_path, and find the first instance of the value (look through column A first, then B, etc).

    If this value does not exist in the tab, will return None. 

    If the value does exist, then this function will walk down the column until it hits the bottom_left_value. 
    If the bottom_left_value is defined but does not exist, will take to the end of the defined row.
    
    Then, will walk down the first row until it hits an empty column. Then it will return the range that 
    defines that rectangular of defined data (with value in the upper left corner).
    """
    workbook = load_workbook(file_path)
    sheet = workbook[sheet_name]

    # We get the last defined rows, so we don't waste time searching data we don't need
    dimension = sheet.calculate_dimension()
    (min_search_col, min_search_row), (max_search_col, max_search_row) = get_col_and_row_indexes_from_range(dimension)
    # Unfortunately, openpyxl indexes from 1, so we add one to treat everything in that range
    (min_search_col, min_search_row), (max_search_col, max_search_row) = (min_search_col + 1, min_search_row + 1), (max_search_col + 1, max_search_row + 1)

    # Loop over the columns one by one to find where this value is set
    min_found_col_index, min_found_row_index = None, None
    for col in sheet.iter_cols(min_row=min_search_row, max_row=max_search_row, min_col=min_search_col, max_col=max_search_col):
        for cell in col:
            if cell.value == upper_left_value:
                min_found_col_index, min_found_row_index = cell.column, cell.row
                break
        
        # As soon as we find something, stop looking
        if min_found_col_index is not None or min_found_row_index is not None:
            break

    if min_found_col_index is None or min_found_row_index is None:
        return None

    column = sheet[get_column_from_column_index(min_found_col_index - 1)] # We need to subtract 1 as we 0 index
    max_found_row_index = None
    for cell in column:
        if cell.row < min_found_row_index:
            continue
        
        # Stop as soon as we match the final value
        if bottom_left_value is None and cell.value is None:
            max_found_row_index = cell.row - 1 # minus b/c this is one past the end
            break
        elif bottom_left_value == cell.value:
            max_found_row_index = cell.row
            break

    # If we looped over the entire column without ending, then we set the max row index
    # as the length of the entire column
    if max_found_row_index is None:
        max_found_row_index = len(column)

    # Then we find find where the rows are defined to
    max_found_col_index = None
    for row in sheet.iter_rows(min_row=min_found_row_index, max_row=min_found_row_index, min_col=min_found_col_index):
        for cell in row:
            if cell.value is None:
                max_found_col_index = cell.column - 1 # minus b/c this is one past the end
                break
    
    # Similarly, if we don't find any empty value in the defined cells, we set the max_col index
    # as the limit of the sheet
    if max_found_col_index is None:
        max_found_col_index = max_search_col

    return f'{get_column_from_column_index(min_found_col_index - 1)}{min_found_row_index}:{get_column_from_column_index(max_found_col_index - 1)}{max_found_row_index}'


def get_read_excel_params_from_range(range: str) -> Tuple[int, int, str]:
    ((start_col_index, start_row_index), (end_col_index, end_row_index)) = get_col_and_row_indexes_from_range(range)
    nrows = end_row_index - start_row_index
    usecols = get_column_from_column_index(start_col_index) + ':' + get_column_from_column_index(end_col_index)
    return start_row_index, nrows, usecols


from datetime import datetime, timedelta
from typing import Optional, Union

import pandas as pd


def get_datetime_format(s: str) -> Optional[str]:
    try:
        # First try letting pandas guess the correct datetime
        converted = pd.to_datetime(s, errors='coerce', infer_datetime_format=True)
        if not pd.isnull(converted):
            return None

        # TODO: Add the most popular formats to here and check them first before 
        # trying all of the formats below for performance.

        FORMATS = [
            '%m{s}%d{s}%Y', 
            '%d{s}%m{s}%Y', 
            '%Y{s}%d{s}%m', 
            '%Y{s}%m{s}%d', 
            '%m{s}%d{s}%Y %H:%M:%S', 
            '%d{s}%m{s}%Y  %H:%M:%S', 
            '%Y{s}%d{s}%m  %H:%M:%S', 
            '%Y{s}%m{s}%d  %H:%M:%S'
        ]
        SEPERATORS = ['/', '-', '.', ',', ':', ' ', '']

        for seperator in SEPERATORS:
            if seperator in s:
                for _format in FORMATS:
                    format = _format.format(s=seperator)
                    result = pd.to_datetime(
                        s,
                        format=format,
                        errors='coerce'
                    )
                    if not pd.isnull(result):
                        return format

        return None
    except:
        return None


def cast_string_to_datetime(
        s: str,
    ) -> Optional[Union[pd.Timestamp, datetime]]:

    # TODO: improve this to work element wise!
    datetime_format = get_datetime_format(s)
    if datetime_format is not None:
        result = pd.to_datetime(
            s,
            format=datetime_format,
            errors='coerce'
        ) 
    else:
        result = pd.to_datetime(
            s,
            infer_datetime_format=True,
            errors='coerce'
        )
    
    # Make sure we don't return a NaT
    if pd.isnull(result):
        return None
    else:
        return result # type: ignore

def cast_to_datetime(unknown: Union[str, int, float, bool, datetime, timedelta]) -> Optional[Union[datetime, pd.Timestamp]]:
    if isinstance(unknown, str):
        return cast_string_to_datetime(unknown)
    elif isinstance(unknown, int):
        return datetime.fromtimestamp(unknown)
    elif isinstance(unknown, float):
        return datetime.fromtimestamp(unknown)
    elif isinstance(unknown, datetime) or isinstance(unknown, pd.Timestamp):
        return unknown

    return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains all functions that can be used in a sheet that operate on
numbers.

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""
from typing import Union
import sys
import numpy as np

import pandas as pd

from mitosheet.public.v3.rolling_range import RollingRange
from mitosheet.public.v3.types.decorators import cast_values_in_arg_to_type

NumberFunctionReturnType = Union[pd.Series, int, float]

@cast_values_in_arg_to_type('number')
def AVG(*argv: Union[int, float, None, pd.Series, RollingRange]) -> NumberFunctionReturnType:

    # Calculate the sum using the SUM function
    sum_for_avg = SUM(*argv)

    # Then, we count the number of entries in the arguements. To match Excel's behavior, we do not count 
    # nan values, and we do not count when a RollingRange runs off the end of the sum
    num_entries: Union[pd.Series,int] = 0

    for arg in argv:
        if isinstance(arg, pd.DataFrame):
            num_non_null_values = arg.count().sum()
            num_entries += num_non_null_values

        elif isinstance(arg, RollingRange):
            num_non_null_values_series = arg.apply(lambda df: df.count().sum())
            num_entries += num_non_null_values_series
            
        elif isinstance(arg, pd.Series):
            # Because series are summed row-wise, we need to count the number of non-null
            # values in the row, which we do with this little type cast
            num_entries += (~arg.isna()).astype(int)
            
        elif arg is not None:
            num_entries += 1

    return sum_for_avg / num_entries if num_entries is not 0 else 0


@cast_values_in_arg_to_type('number')
def MAX(*argv: Union[int, float, None, pd.Series, RollingRange]) -> NumberFunctionReturnType:
    
    result: Union[pd.Series, float, int] = -sys.maxsize - 1

    def get_new_result(_result: Union[pd.Series, float, int], new_value: Union[pd.Series, float, int]) -> Union[pd.Series, float, int]:
        if isinstance(_result, pd.Series):
            if isinstance(new_value, pd.Series):
                return pd.concat([_result, new_value], axis=1).max(axis=1)
            else:
                return _result.apply(lambda v: np.nanmax([v, new_value]))
        else:
            if isinstance(new_value, pd.Series):
                return new_value.apply(lambda v: np.nanmax([v, _result]))
            else:
                return max(_result, new_value)

    for arg in argv:
        if isinstance(arg, pd.DataFrame):
            df_sum = arg.sum().sum()
            result = get_new_result(result, df_sum)

        elif isinstance(arg, RollingRange):
            new_series = arg.apply(lambda df: df.sum().sum())
            result = get_new_result(result, new_series)
            
        elif isinstance(arg, pd.Series):
            result = get_new_result(result, arg)
            
        elif arg is not None:
            result = get_new_result(result, arg)

    # If we don't find any arguements, we default to 0 -- like Excel
    kept_default_min_value = not isinstance(result, pd.Series) and result == (-sys.maxsize - 1)
    return result if not kept_default_min_value else 0

@cast_values_in_arg_to_type('number')
def SUM(*argv: Union[int, float, None, pd.Series, RollingRange]) -> NumberFunctionReturnType:
    
    result: Union[pd.Series, float, int] = 0

    for arg in argv:

        if isinstance(arg, pd.DataFrame):
            result += arg.sum().sum()
        elif isinstance(arg, RollingRange):
            result += arg.apply(lambda df: df.sum().sum())
        elif isinstance(arg, pd.Series):
            result += arg.fillna(0)
        elif arg is not None:
            result += arg

    return result 



# TODO: we should see if we can list these automatically!
NUMBER_FUNCTIONS = {
    'AVG': AVG,
    'MAX': MAX,
    'SUM': SUM,
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
from typing import Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.excel_utils import (get_column_from_column_index,
                                   get_col_and_row_indexes_from_range)
from mitosheet.public.v2.excel_utils import get_read_excel_params_from_range
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import column_header_to_transpiled_code
from mitosheet.types import ExcelRangeImport


EXCEL_RANGE_IMPORT_TYPE_RANGE = 'range'
EXCEL_RANGE_IMPORT_TYPE_UPPER_LEFT_VALUE = 'upper left corner value'
EXCEL_RANGE_END_CONDITION_FIRST_EMPTY_VALUE = 'first empty cell'
EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE = 'bottom left corner value'


class ExcelRangeImportCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, file_path: str, sheet_name: str, range_imports: List[ExcelRangeImport]):
        super().__init__(prev_state, post_state)
        self.file_path = file_path
        self.sheet_name = sheet_name
        self.range_imports = range_imports

    def get_display_name(self) -> str:
        return 'Excel Range Import'
    
    def get_description_comment(self) -> str:
        
        return f"Imported {len(self.range_imports)} dataframes from {self.sheet_name} in {self.file_path}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        code = []
        for idx, range_import in enumerate(self.range_imports):

            sheet_index = len(self.prev_state.dfs) + idx
            df_name = self.post_state.df_names[sheet_index]

            # If it's an explicit range, then just import that exact range
            if range_import['type'] == EXCEL_RANGE_IMPORT_TYPE_RANGE:
                _range = range_import['value']
                skiprows, nrows, usecols = get_read_excel_params_from_range(_range)
                
                code.append(
                    f'{df_name} = pd.read_excel(\'{self.file_path}\', sheet_name=\'{self.sheet_name}\', skiprows={skiprows}, nrows={nrows}, usecols=\'{usecols}\')'
                )

            else:
                # Otherwise, if you're importing based on values, we generate dynamic code

                end_condition = range_import['end_condition'] # type: ignore

                if end_condition['type'] == EXCEL_RANGE_END_CONDITION_FIRST_EMPTY_VALUE:
                    code.extend([
                        f'_range = get_table_range_from_upper_left_corner_value({column_header_to_transpiled_code(self.file_path)}, {column_header_to_transpiled_code(self.sheet_name)}, {column_header_to_transpiled_code(range_import["value"])})',
                        'skiprows, nrows, usecols = get_read_excel_params_from_range(_range)'
                    ])
                elif end_condition['type'] == EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE:
                    code.extend([
                        f'_range = get_table_range_from_upper_left_corner_value({column_header_to_transpiled_code(self.file_path)}, {column_header_to_transpiled_code(self.sheet_name)}, {column_header_to_transpiled_code(range_import["value"])}, {column_header_to_transpiled_code(end_condition["value"])})',
                        'skiprows, nrows, usecols = get_read_excel_params_from_range(_range)'
                    ])
                else:
                    raise Exception(f"End condition {end_condition} is not a valid type.")
                
                code.append(
                    f'{df_name} = pd.read_excel(\'{self.file_path}\', sheet_name=\'{self.sheet_name}\', skiprows=skiprows, nrows=nrows, usecols=usecols)'
                )

            # Add a new line in between different imports otherwise the code looks bad
            if idx < len(self.range_imports) - 1:
                code.append('')

        return code, ['import pandas as pd']

    def get_created_sheet_indexes(self) -> Optional[List[int]]:
        return [i for i in range(len(self.post_state.dfs) - len(self.range_imports), len(self.post_state.dfs))]

    def _combine_right_with_excel_range_import_code_chunk(self, excel_range_import_code_chunk: "ExcelRangeImportCodeChunk") -> Optional[CodeChunk]:
        if excel_range_import_code_chunk.file_path == self.file_path and excel_range_import_code_chunk.sheet_name == self.sheet_name:
            new_range_imports = copy(self.range_imports)
            new_range_imports.extend(excel_range_import_code_chunk.range_imports)

            return ExcelRangeImportCodeChunk(
                self.prev_state,
                excel_range_import_code_chunk.post_state,
                self.file_path,
                self.sheet_name,
                new_range_imports
            )

        return None


    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, ExcelRangeImportCodeChunk):
            return self._combine_right_with_excel_range_import_code_chunk(other_code_chunk)

        return None
    








from datetime import datetime, timedelta
from functools import wraps
from typing import Any, Callable

import pandas as pd

from mitosheet.public.v3.rolling_range import RollingRange
from mitosheet.public.v3.types.utils import CONVERSION_FUNCTIONS
from mitosheet.types import PrimitiveTypeName


def is_primitive_value(value: Any) -> bool:
    return isinstance(value, str) or \
        isinstance(value, int) or \
        isinstance(value, float) or \
        isinstance(value, bool) or \
        isinstance(value, datetime) or \
        isinstance(value, timedelta)
        # TODO: nan here?

def cast_values_in_arg_to_type(
    target_primitive_type_name: PrimitiveTypeName,
) -> Callable:

    def wrap(sheet_function):
        @wraps(sheet_function)
        def wrapped_sheet_function(*args):   

            final_args = []
            conversion_function = CONVERSION_FUNCTIONS[target_primitive_type_name]

            # For every arguement, go through and cast them to the correct type
            for arg in args:

                if is_primitive_value(arg):
                    final_args.append(conversion_function(arg))
                
                elif isinstance(arg, pd.Series):
                    final_args.append(arg.apply(conversion_function)) # Why doesn't type work?

                elif isinstance(arg, pd.DataFrame):
                    final_args.append(arg.apply(lambda c: c.apply(conversion_function)))

                elif isinstance(arg, RollingRange):
                    obj = arg.obj
                    new_obj = obj.apply(lambda c: c.apply(conversion_function))
                    final_args.append(RollingRange(new_obj, arg.window, arg.offset))
                
            return sheet_function(*final_args)        
        return wrapped_sheet_function
    return wrap



from mitosheet.types import ColumnHeader


def flatten_column_header(column_header: ColumnHeader) -> ColumnHeader:
    """
    Given a pandas column header, if it is a list or tuple, it will
    flatten this header into a string. 

    This is useful for headers that are created from a pivot table,
    where they are often tuples or lists.

    NOTE: This function, if applied before the depricated make_valid_header
    function, should return the same result for any input. That is:
    For all column headers:
    make_valid_header(flatten_column_header(ch)) = make_valid_header(ch)
    """
    if isinstance(column_header, list) or isinstance(column_header, tuple):
        column_header_str = [str(ch) for ch in column_header]
        return ' '.join(column_header_str).strip()

    return column_header